"""
Tile package reading functions.
see info here: https://gdbgeek.wordpress.com/2012/08/09/demystifying-the-esri-compact-cache/

Tile package files:
*.bundlx: tile index, tile offsets in bundle are stored in 5 byte values
*.bundle: offsets are stored in bundlx, first 4 bytes at offset are length of tile data
conf.xml: basic tileset info
conf.cdi: tileset bounding box
"""


import os
import math
import time
from xml.etree import ElementTree
from collections import namedtuple
from zipfile import ZipFile
from io import BytesIO
import logging
from tpkutils.mbtiles import Mbtiles
from tpkutils.util import geo_bounds


logger = logging.getLogger('tpkutils')
# logger.setLevel(logging.DEBUG)
# logger.addHandler(logging.StreamHandler())


BUNDLE_DIM = 128 # bundles are 128 rows x 128 columns tiles
# TODO: bundle size is stored in one of the configuration files
INDEX_SIZE = 5  # tile index is stored in 5 byte values


Tile = namedtuple('Tile', ['z', 'x', 'y', 'data'])


def buffer_to_offset(buffer):
    """
    Convert a byte buffer into an integer offset according to ArcGIS packing
    scheme:

    (buffer[0] & 0xff) +
    (buffer[1] & 0xff) * 2 ** 8 +
    (buffer[2] & 0xff) * 2 ** 16 +
    (buffer[3] & 0xff) * 2 ** 24 ...

    Parameters
    ----------
    buffer: list[bytes]
        byte buffer

    Returns
    -------
    int: offset
    """

    return sum(((v & 0xff) * 2 ** (i * 8) for i, v in enumerate(buffer)))


def read_tile(bundle, offset):
    """
    Read tile bytes at offset position in bundle.

    Parameters
    ----------
    bundle: ArcGIS tile bundle file
    offset: offset in bytes to beginning of tile data block (first 4 bytes are length)

    Returns
    -------
    tile_bytes: bytes (may be empty)
    """

    bundle.seek(offset)
    return bundle.read(buffer_to_offset(bundle.read(4)))






# TODO: context manager
class TPK(object):
    def __init__(self, filename):
        """
        Opens a tile package file for reading tiles and metadata.

        Parameters
        ----------
        filename: string
            name of tile package
        """
        self._fp = ZipFile(filename)

        # Fields specifically meant to be updated by user
        self.version = '1.0.0'
        self.attribution = ''

        logger.debug('Reading package metadata')
        tile_root = 'v101/Layers'  # TODO: automatically determine

        # Bounding box and projection info is in .../Layers/conf.cdi
        conf_filename = '{0}/{1}'.format(tile_root, 'conf.cdi')
        xml = ElementTree.fromstring(self._fp.read(conf_filename))
        wm_bounds = [
            float(xml.find(e).text) for e in ('XMin', 'YMin', 'XMax', 'YMax')
        ]
        self.bounds = geo_bounds(*wm_bounds)

        # File format, zoom levels, etc in .../Layers/conf.xml
        conf_filename = '{0}/{1}'.format(tile_root, 'conf.xml')
        xml = ElementTree.fromstring(self._fp.read(conf_filename))
        self.zoom_levels = [
            int(e.text) for e in
            xml.findall('TileCacheInfo/LODInfos/LODInfo/LevelID')
        ]
        self.format = xml.find('TileImageInfo/CacheTileFormat').text

        # Descriptive info in esriinfo/iteminfo.xml
        # Some fields are required by ArcGIS to create tile package
        xml = ElementTree.fromstring(self._fp.read('esriinfo/iteminfo.xml'))
        self.name = xml.find('title').text  # required field, provided automatically
        self.summary = xml.find('summary').text  # required field
        self.tags = xml.find('tags').text or ''  # required field
        self.description = xml.find('description').text or ''  # optional
        self.credits = xml.find('accessinformation').text or ''  # optional, Credits in ArcGIS
        self.use_constraints = xml.find('licenseinfo').text or ''  # optional, Use Constraints in ArcGIS

    def read_tiles(self, zoom=None, flip_y=False):
        """
        Read all non-empty tiles from tile package, optionally limited to zoom
        levels provided.

        Parameters
        ----------
        zoom: int or list-like  (default: None)
            zoom level or list-like of zoom levels
        flip_y: bool  (default False)
            if True, will return tiles in xyz tile scheme.  Otherwise will use
            ArcGIS scheme.

        Returns
        -------
        generator(Tile)
            Only returns non-empty tiles
        """

        if isinstance(zoom, int):
            zoom = [zoom]

        bundles = []
        for name in self._fp.namelist():
            if 'Layers/_alllayers/L' in name and '.bundle' in name:
                z = int(name.split('/')[-2].lstrip('L'))
                if zoom is None or z in zoom:
                    bundles.append(name)

        for fname in bundles:
            logger.debug('Reading bundle: {0}'.format(fname))
            start = time.time()
            # parse filename to determine row / col offset for bundle
            # offsets are in hex
            root = os.path.splitext(os.path.basename(fname))[0]
            r_off, c_off = [int(x, 16) for x in root.lstrip('R').split('C')]
            # zoom level is from name of containing folder
            z = int(os.path.split(os.path.dirname(fname))[1].lstrip('L'))

            # discard 16 byte header
            index_bytes = self._fp.read(fname.replace('.bundle', '.bundlx'))[16:]

            bundle_bytes = BytesIO(self._fp.read(fname))
            index = 0
            max_index = BUNDLE_DIM ** 2
            while index < max_index:
                data = read_tile(
                    bundle_bytes,
                    buffer_to_offset(
                        index_bytes[index * INDEX_SIZE:(index + 1) * INDEX_SIZE]
                    )
                )
                if data:
                    row = math.floor(float(index) / BUNDLE_DIM)
                    # Note: x and y seem backwards but were verified through trial and error!
                    x = c_off + row
                    y = r_off + index - (row * BUNDLE_DIM)

                    if flip_y:
                        y = (1 << z) - 1 - y

                    yield Tile(z, x, y, data)

                index += 1
            logger.debug('Time to read bundle: {0:2f}'.format(time.time() - start))

    def to_mbtiles(self, filename, zoom=None, overwrite=False):
        """
        Export tile package to mbtiles v1.1 file, optionally limited to zoom
        levels.

        Parameters
        ----------
        filename: string
            name of mbtiles file
        zoom: int or list-like of ints, default: None (all tiles exported)
            zoom levels to export to mbtiles
        overwrite: bool, default: False
            overwrite existing mbtiles file.  If False and file exists, an
            exception will be raised.
        """

        if self.format.lower() == 'mixed':
            raise ValueError('Mixed format tiles are not supported for export to mbtiles')

        mbtiles = Mbtiles(filename, overwrite)
        if zoom is None:
            zoom = self.zoom_levels
        elif isinstance(zoom, int):
            zoom = [zoom]

        zoom = list(zoom)
        zoom.sort()

        for tile in self.read_tiles(zoom, flip_y=True):
            mbtiles.add_tile(tile.z, tile.x, tile.y, tile.data)

        bounds = self.bounds
        center = '{0:4f},{1:4f},{2}'.format(
            bounds[0] + (bounds[2] - bounds[0]) / 2.0,
            bounds[1] + (bounds[3] - bounds[1]) / 2.0,
            max(zoom[0], int((zoom[-1] - zoom[0]) / 4.0))  # Tune this
        )

        mbtiles.set_metadata({
            'name': self.name,
            'description': self.summary,  # not description, which is optional
            'version': self.version,
            'attribution': self.attribution,
            'tags': self.tags,
            'credits': self.credits,
            'use_constraints': self.use_constraints,

            'type': 'overlay',
            'format': self.format.lower().replace('jpeg', 'jpg')[:3],
            'bounds': ','.join('{0:4f}'.format(v) for v in self.bounds),
            'center': center,
            'minzoom': str(zoom[0]),
            'maxzoom': str(zoom[-1]),

            'legend': ''  # TODO: extract from svc symbols
        })

        mbtiles.close()

    def close(self):
        self._fp.close()